/**
  * File: CoordinationsMap.java
  * Content: Wrapper over a map of coordinations, including a boolean value to show when it is finished
  *          every coordinations. Each coordination is a SubgroupCoordination, and the map has a key the
  *          subgroup id. It is considered to have valid coordinations when some coordination is not
  *          null and has not set transferFinished
  * Author: LuisM Pena
  * Date: 28th June 2001
  * Version: 0.51.00
  * Last change:
  *
  *
  **/

package sensei.domains;

import sensei.middleware.domains.SubgroupCoordination;
import sensei.middleware.domains.PhaseCoordination;

import sensei.util.IntWrapper;

import java.util.Iterator;
import java.util.Map;

/**
  * Handle any operation to be performed on subgroups
  **/
class CoordinationsMap {
	public CoordinationsMap(Map coordinations) {
		Trace.code("CoordinationsMap.java -> CoordinationsMap ( Map coordinations )");
		this.coordinationsMap = coordinations;
		this.anyValidCoordination = getMapState();
	}
	public CoordinationsMap(Map coordinations, boolean anyValidCoordination) {
		Trace.code("CoordinationsMap.java -> CoordinationsMap ( Map coordinations , boolean anyValidCoordination )");
		this.coordinationsMap = coordinations;
		this.anyValidCoordination = anyValidCoordination;
	}

	public void freeResources() {
		Trace.code("CoordinationsMap.java -> void freeResources ( )");
		if (coordinationsMap != null) {
			reset();
			coordinationsMap.clear();
			coordinationsMap = null;
		}
	}

	public PhaseCoordination getPhase(IntWrapper subgroup) {
		Trace.code("CoordinationsMap.java -> PhaseCoordination getPhase ( IntWrapper subgroup )");
		SubgroupCoordination temp = (SubgroupCoordination) coordinationsMap.get(subgroup);
		return temp == null ? null : temp.phase;
	}
	public boolean isValidSubgroupCoordination(PhaseCoordination phase) {
		Trace.code("CoordinationsMap.java -> boolean isValidSubgroupCoordination ( PhaseCoordination phase )");
		return (phase != null) && (!phase.transferFinished);
	}
	public void resetValidity() {
		Trace.code("CoordinationsMap.java -> void resetValidity ( )");
		anyValidCoordination = false;
	}
	public void updateValidity(PhaseCoordination phase) {
		Trace.code("CoordinationsMap.java -> void updateValidity ( PhaseCoordination phase )");
		anyValidCoordination = anyValidCoordination || isValidSubgroupCoordination(phase);
	}
	/**
	 * Returns true if the coordinations map contains some valid phase
	 */
	public boolean getMapState() {
		Trace.code("CoordinationsMap.java -> boolean getMapState ( )");
		boolean ret = false;
		Iterator it = coordinationsMap.entrySet().iterator();
		while (it.hasNext() && !ret) {
			Map.Entry entry = (Map.Entry) (it.next());
			SubgroupCoordination subgroupCoordination = (SubgroupCoordination) (entry.getValue());
			ret = isValidSubgroupCoordination(subgroupCoordination.phase);
		}
		return ret;
	}
	/**
	 * Put all the phases to null, resetting the validity as well
	 */
	public void reset() {
		Trace.code("CoordinationsMap.java -> void reset ( )");
		Iterator it = coordinationsMap.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) (it.next());
			SubgroupCoordination subgroupCoordination = (SubgroupCoordination) (entry.getValue());
			subgroupCoordination.phase = null;
		}
		resetValidity();
	}
	public Map coordinationsMap; //IntWrapper, SubgroupCoordination
	public boolean anyValidCoordination;
}
