/**
	* File: SubgroupsManager.java
	* Content: Handler of subgroups: how to call them, adding new subgroups ...
	* Author: LuisM Pena
	* Date: 28th June 2001
	* Version: 0.51.00
	* Last change:
	*
	**/

package sensei.domains;

import sensei.middleware.gms.GMSnarrower;
import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.GroupMember;
import sensei.middleware.gms.Message;
import sensei.middleware.gms.ObjectsHandling;
import sensei.middleware.gms.View;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.SubgroupCoordination;
import sensei.middleware.domains.SubgroupsHandlerException;
import sensei.middleware.domains.SubgroupsHandlerExceptionReason;
import sensei.middleware.domains.SubgroupProperties;

import sensei.util.IntWrapper;

import java.io.IOException;
import java.util.Iterator;
import java.util.HashMap;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;

/**
	* Handle any operation to be performed on subgroups
	**/
class SubgroupsManager extends SubgroupsTasks {

	//*************************************************************************************//
	//**************************** CONSTRUCTOR ********************************************//
	//*************************************************************************************//

	public SubgroupsManager() {
		Trace.code("SubgroupsManager.java -> SubgroupsManager ( )");
		subgroups = new TreeMap();
		existingGroups = new HashMap();
		subgroupProperties = null;
		singleArray = new int[1];
		wrapper = new IntWrapper();
	}

	//*************************************************************************************//
	//**************************** FREE RESOURCES *****************************************//
	//*************************************************************************************//

	/**
	 * FreeResources releases any resources, leaving the object in unusable state
	 */
	public synchronized void freeResources() {
		Trace.code("SubgroupsManager.java -> synchronized void freeResources ( )");
		if (subgroups != null) {
			clear();
			super.freeResources();
			subgroups = null;
			singleArray = null;
			wrapper = null;
			existingGroups.clear();
			existingGroups = null;
		}
	}

	//*************************************************************************************//
	//**************************** CLEAR **************************************************//
	//*************************************************************************************//

	/**
	 * Clear removes some of the information, but lets the manager in initial state
	 */
	public synchronized void clear() {
		Trace.code("SubgroupsManager.java -> synchronized void clear ( )");
		subgroupProperties = null;
		Iterator it = subgroups.values().iterator();
		while (it.hasNext())
			freeResourcesOnSubgroup((SubgroupInfo) it.next());
		subgroups.clear();
	}

	//*************************************************************************************//
	//**************************** CONTAINS SUBGROUP **************************************//
	//*************************************************************************************//

	/**
	 * Returns true if the given subgroup exists
	 */
	public synchronized boolean containsSubgroup(int subgroup) {
		Trace.code("SubgroupsManager.java -> synchronized boolean containsSubgroup ( int subgroup )");
		wrapper.i = subgroup;
		return subgroups.containsKey(wrapper);
	}

	//*************************************************************************************//
	//**************************** CHECK STATIC SUBGROUPS *********************************//
	//*************************************************************************************//

	/**
	 * Verifies the subgroups composition, checking that the static subgroups exist, returning false
	 * otherwise
	 */
	public synchronized boolean checkStaticSubgroups(SubgroupProperties subgroupProperties[]) {
		Trace.code("SubgroupsManager.java -> synchronized boolean checkStaticSubgroups ( SubgroupProperties subgroupProperties [ ] )");
		boolean ret = true;
		int i = subgroupProperties.length;
		while (ret && (i-- > 0)) {
			SubgroupProperties subgroup = subgroupProperties[i];
			if (subgroup.id <= Consts.MAX_STATIC_SUBGROUP_ID) {
				wrapper.i = subgroup.id;
				SubgroupInfo info = (SubgroupInfo) subgroups.get(wrapper);
				if ((info == null) || (info.subgroup.getStateTransferType() != subgroup.type))
					ret = false;
			}
		}
		return ret;
	}

	//*************************************************************************************//
	//**************************** GET SUBGROUP PROPERTIES ********************************//
	//*************************************************************************************//

	/**
		* Obtains the properties of the subgroups
		**/
	public synchronized SubgroupProperties[] getSubgroupProperties() {
		Trace.code("SubgroupsManager.java -> synchronized SubgroupProperties [ ] getSubgroupProperties ( )");
		if (subgroupProperties == null) {
			int howmany = subgroups.size();
			subgroupProperties = new SubgroupProperties[howmany];
			if (howmany > 0) {
				int index = 0;
				Iterator iter = subgroups.entrySet().iterator();
				while (iter.hasNext()) {
					Map.Entry entry = (Map.Entry) (iter.next());
					IntWrapper subgroup = (IntWrapper) (entry.getKey());
					SubgroupInfo subgroupInfo = (SubgroupInfo) (entry.getValue());
					subgroupProperties[index++] = new SubgroupProperties(subgroup.i, subgroupInfo.info, subgroupInfo.subgroup.getStateTransferType());
				}
				assert index == howmany;
			}
		}
		return subgroupProperties;
	}

	//*************************************************************************************//
	//**************************** GET SUBGROUPS ******************************************//
	//*************************************************************************************//

	/**
		* Returns the list of subgroups
		**/
	public synchronized int[] getSubgroups() {
		Trace.code("SubgroupsManager.java -> synchronized int [ ] getSubgroups ( )");
		int howmany = subgroups.size();
		int ret[] = new int[howmany];
		if (howmany > 0) {
			int index = 0;
			Iterator iter = subgroups.keySet().iterator();
			while (iter.hasNext()) {
				ret[index++] = ((IntWrapper) (iter.next())).i;
			}
			assert index == howmany;
		}
		return ret;
	}

	/**
		* Returns the subgroup associated to a given subgroupId, null if not existing
		*/
	GroupMember getSubgroup(int id) {
		Trace.code("SubgroupsManager.java -> GroupMember getSubgroup ( int id )");
		wrapper.i = id;
		SubgroupInfo info = (SubgroupInfo) subgroups.get(wrapper);
		return info == null ? null : info.subgroup.getGroupMember();
	}

	//	 /**
	//		* Returns the subgroupId for a group. If the group is not valid, EVERY_SUBGROUP is returned
	//		*/
	//	public int getSubgroupId(GroupMember subgroup)
	//	{
	//		Trace.code("SubgroupsManager.java -> int getSubgroupId(GroupMember subgroup)");
	//		Object obj = existingGroups.get(subgroup);
	//		return obj==null? Consts.EVERY_SUBGROUP : ((IntWrapper)obj).i;
	//	}
	//
	/**
		* Returns the subgroupId for a group. If the group is not valid, EVERY_SUBGROUP is returned
		*/
	public synchronized int getSubgroupId(GroupMember subgroup) {
		Trace.code("SubgroupsManager.java -> synchronized int getSubgroupId ( GroupMember subgroup )");
		Object obj = existingGroups.get(subgroup);
		if (obj == null) {
			return searchSubgroupId(subgroup);
		}
		return ((IntWrapper) obj).i;
	}

	int searchSubgroupId(GroupMember subgroup) {
		Trace.code("SubgroupsManager.java -> int searchSubgroupId ( GroupMember subgroup )");
		for (Iterator it = existingGroups.entrySet().iterator(); it.hasNext();) {
			Map.Entry entry = (Map.Entry) it.next();
			if (ObjectsHandling.areEquivalent(subgroup, GMSnarrower.toGroupMember(entry.getKey()))) {
				return ((IntWrapper) entry.getValue()).i;
			}
		}
		return Consts.EVERY_SUBGROUP;
	}

	//*************************************************************************************//
	//**************************** OPERATIONS TO MODIFY THE SUBGROUPS *********************//
	//*************************************************************************************//

	/**
	 * @exception SubgroupException can only be sent in case of dynamic groups
	 */
	public void registerSubgroup(int uniqueSubgroupId, STGroupMember subgroup, DynamicSubgroupInfo info) throws SubgroupsHandlerException, SubgroupException {
		Trace.code("SubgroupsManager.java -> void registerSubgroup ( int uniqueSubgroupId , STGroupMember subgroup , DynamicSubgroupInfo info )");
		registerSubgroup(new IntWrapper(uniqueSubgroupId), subgroup, info);
	}

	public synchronized void registerSubgroup(IntWrapper uniqueSubgroupId, STGroupMember subgroup, DynamicSubgroupInfo info) throws SubgroupsHandlerException, SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void registerSubgroup ( IntWrapper uniqueSubgroupId , STGroupMember subgroup , DynamicSubgroupInfo info )");
		if (subgroups.containsKey(uniqueSubgroupId))
			throw new SubgroupsHandlerException(SubgroupsHandlerExceptionReason.SubgroupIdAlreadyInUse);

		subgroups.put(uniqueSubgroupId, new SubgroupInfo(subgroup, info));
		existingGroups.put(subgroup.getGroupMember(), uniqueSubgroupId);
		subgroupProperties = null; //the cache must be restarted!

		//if memberAccpeted has already been sent to other members, send memberAccepted to this group
		if (postMemberAcceptanceTask != null) {
			try {
				postMemberAcceptanceTask.call(subgroup);
			}
			catch (IOException ex) {
				throw new SubgroupException(uniqueSubgroupId.i, ex);
			}
		}
	}

	public synchronized void unregisterSubgroup(int uniqueSubgroupId) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void unregisterSubgroup ( int uniqueSubgroupId )");
		wrapper.i = uniqueSubgroupId;
		SubgroupInfo groupMember = (SubgroupInfo) subgroups.remove(wrapper);

		if (groupMember != null) {
			existingGroups.remove(groupMember.subgroup.getGroupMember());
			//send memberExcluded to this group
			try {
				postMemberExpulsionTask.call(groupMember.subgroup);
			}
			catch (IOException ex) {
				throw new SubgroupException(uniqueSubgroupId, ex);
			}
			catch (RuntimeException ex) {
				throw new SubgroupException(uniqueSubgroupId, ex);
			}

			freeResourcesOnSubgroup(groupMember);
			subgroupProperties = null; //the cache must be restarted!
		}
	}

	void freeResourcesOnSubgroup(SubgroupInfo subgroupInfo) {
		Trace.code("SubgroupsManager.java -> void freeResourcesOnSubgroup ( SubgroupInfo subgroupInfo )");
		if (subgroupInfo != null) {
			subgroupInfo.info = null;
			subgroupInfo.subgroup.freeResources();
			subgroupInfo.subgroup = null;
		}
	}

	//*************************************************************************************//
	//**************************** INIT COORDINATIONS *************************************//
	//*************************************************************************************//

	/**
	 * Initialites the phases to null values
	 */
	public synchronized CoordinationsMap initCoordinations() throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized CoordinationsMap initCoordinations ( )");
		Map ret = new HashMap(subgroups.size());
		Iterator it = subgroups.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) (it.next());
			IntWrapper subgroup = (IntWrapper) (entry.getKey());
			SubgroupInfo subgroupInfo = (SubgroupInfo) (entry.getValue());
			if (needsTransfer(subgroup.i, subgroupInfo))
				ret.put(subgroup, new SubgroupCoordination(subgroup.i, null));
		}
		return new CoordinationsMap(ret, false);
	}

	boolean needsTransfer(int id, SubgroupInfo subgroupInfo) {
		Trace.code("SubgroupsManager.java -> boolean needsTransfer ( int id , SubgroupInfo subgroupInfo )");
		return subgroupInfo.subgroup.needsTransfer();
	}

	//*************************************************************************************//
	//**************************** CHECK DYNAMIC COORDINATIONS ****************************//
	//*************************************************************************************//

	/**
	 * Checks the coordinations, to add new dynamic groups and remove the removed groups
	 */
	public synchronized CoordinationsMap checkDynamicCoordinations(CoordinationsMap previous) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized CoordinationsMap checkDynamicCoordinations ( CoordinationsMap previous )");

		Map ret = new HashMap(subgroups.size());
		Iterator it = subgroups.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) (it.next());
			IntWrapper subgroup = (IntWrapper) (entry.getKey());
			SubgroupInfo subgroupInfo = (SubgroupInfo) (entry.getValue());
			if (needsTransfer(subgroup.i, subgroupInfo))
				ret.put(subgroup, new SubgroupCoordination(subgroup.i, previous.getPhase(subgroup)));
		}
		return new CoordinationsMap(ret, false);
	}

	//*************************************************************************************//
	//**************************** OPERATIONS TO CALL THE SUBGROUPS ***********************//
	//*************************************************************************************//

	public synchronized void memberExpulsed() throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void memberExpulsed ( )");
		callSubgroups(new PostMemberExpulsionTask());
	}

	public synchronized void memberAccepted(int identity, GroupHandler handler, View view) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void memberAccepted ( int identity , GroupHandler handler , View view )");
		assert postMemberAcceptanceTask == null;
		postMemberAcceptanceTask = new PostMemberAcceptanceTask(identity, handler, view);
		callSubgroups(postMemberAcceptanceTask);
	}

	public synchronized void changingView() throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void changingView ( )");
		callSubgroups(changeViewTask);
	}

	public synchronized void installView(View view) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void installView ( View view )");
		installViewTask.set(view);
		postMemberAcceptanceTask.set(view); //for dynamic groups
		callSubgroups(installViewTask);
	}

	public synchronized void sendMessage(int sender, Message message) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void sendMessage ( int sender , Message message )");
		sendMessageTask.set(sender, message);
		callSubgroups(sendMessageTask);
	}

	public synchronized void sendMessage(int subgroup, int sender, Message message) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void sendMessage ( int subgroup , int sender , Message message )");
		sendMessageTask.set(sender, message);
		callSubgroup(subgroup, sendMessageTask);
	}

	public synchronized void castMessage(int sender, Message message) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void castMessage ( int sender , Message message )");
		castMessageTask.set(sender, message);
		callSubgroups(castMessageTask);
	}

	public synchronized void castMessage(int subgroup, int sender, Message message) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void castMessage ( int subgroup , int sender , Message message )");
		castMessageTask.set(sender, message);
		callSubgroup(subgroup, castMessageTask);
	}

	/**
	 * Coordinations are modified containing at the end the answers of the members.
	 */
	public synchronized void startTransfer(int target, CoordinationsMap coordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void startTransfer ( int target , CoordinationsMap coordinations )");
		startTransferTask.set(target);
		callSubgroups(startTransferTask, coordinations, true); //it is said to every subgroup
	}

	/**
	 * Informs to every subgroup that the transfer is assumed
	 */
	public synchronized void assumeState() throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void assumeState ( )");
		callSubgroups(new AssumeStateTask()); //it is said to every subgroup
	}

	/**
	 * Informs to every subgroup that the transfer has finished
	 */
	public synchronized void stopTransfer(int joiningMember, CoordinationsMap coordinations, boolean transferFinished) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void stopTransfer ( int joiningMember , CoordinationsMap coordinations , boolean transferFinished )");
		singleArray[0] = joiningMember;
		stopTransfer(singleArray, coordinations, transferFinished);
	}

	/**
	 * Informs to every subgroup that the transfer has finished
	 */
	public synchronized void stopTransfer(int joiningMembers[], CoordinationsMap coordinations, boolean transferFinished) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void stopTransfer ( int joiningMembers [ ] , CoordinationsMap coordinations , boolean transferFinished )");
		assert coordinations != null;
		stopTransferTask.set(joiningMembers, transferFinished);
		callSubgroups(stopTransferTask, coordinations, true); //it is said to every subgroup
	}

	/**
	 * Informs to every subgroup that the transfer must be synchronized
	 */
	public synchronized void syncTransfer(int coordinator, CoordinationsMap coordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void syncTransfer ( int coordinator , CoordinationsMap coordinations )");
		syncTransferTask.set(coordinator);
		callSubgroups(syncTransferTask, coordinations, true); //it is said to every subgroup
	}

	/**
	 * Informs to the transfering groups about the interruption on the transfer
	 */
	public synchronized void interruptTransfer(CoordinationsMap coordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void interruptTransfer ( CoordinationsMap coordinations )");
		callSubgroups(interruptTransferTask, coordinations, false); //it is said to every subgroup still on transfer
	}

	/**
	 * Informs to every subgroup about the continuation on the transfer
	 */
	public synchronized void continueTransfer(int joiningMember, CoordinationsMap coordinations, Map joiningCoordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void continueTransfer ( int joiningMember , CoordinationsMap coordinations , Map joiningCoordinations )");
		singleArray[0] = joiningMember;
		continueTransfer(singleArray, coordinations, joiningCoordinations);
	}

	/**
	 * Informs to the transfering groups about the continuation on the transfer
	 */
	public synchronized void continueTransfer(int joinings[], CoordinationsMap coordinations, Map joiningCoordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void continueTransfer ( int joinings [ ] , CoordinationsMap coordinations , Map joiningCoordinations )");
		continueTransferTask.set(joinings, joiningCoordinations);
		callSubgroups(continueTransferTask, coordinations, true); //it is said to every subgroup
	}

	/**
	 * Requests to the subgroups their state
	 */
	public synchronized Map getState(CoordinationsMap coordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized Map getState ( CoordinationsMap coordinations )");
		getStateTask.reset();
		callSubgroups(getStateTask, coordinations, false); //it is said only to subgroups still on transfer
		return getStateTask.getState();
	}

	/**
	 * Requests to the subgroups their state
	 */
	public synchronized void sendState(Map states, CoordinationsMap coordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> synchronized void sendState ( Map states , CoordinationsMap coordinations )");
		sendStateTask.setState(states);
		//the coordinations can already contain transfers finished, but that specific transfer must still be
		//donem if the states contains information for that subgroup. For that reason, the call is done over
		//every subgroup, and the task will be only done over the members requiring state transfer
		callSubgroups(sendStateTask, coordinations, true);
	}

	//*************************************************************************************//
	//**************************** PRIVATE OPERATIONS *************************************//
	//*************************************************************************************//

	/**
	 * Calls to the specified subgroup with the associated task
	 */
	void callSubgroup(int subgroup, Task task) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> void callSubgroup ( int subgroup , Task task )");
		wrapper.i = subgroup;
		Object theSubgroup = subgroups.get(wrapper);
		if (theSubgroup != null)
			try {
				task.call(((SubgroupInfo) theSubgroup).subgroup);
			}
			catch (IOException ex) {
				throw new SubgroupException(subgroup, ex);
			}
			catch (RuntimeException ex) {
				throw new SubgroupException(subgroup, ex);
			}
	}

	/**
	 * Calls to every subgroup with the associated task
	 */
	void callSubgroups(Task task) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> void callSubgroups ( Task task )");
		Iterator it = subgroups.entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry entry = (Map.Entry) (it.next());
			try {
				task.call(((SubgroupInfo) entry.getValue()).subgroup);
			}
			catch (IOException ex) {
				IntWrapper subgroup = (IntWrapper) (entry.getKey());
				throw new SubgroupException(subgroup.i, ex);
			}
			catch (RuntimeException ex) {
				IntWrapper subgroup = (IntWrapper) (entry.getKey());
				throw new SubgroupException(subgroup.i, ex);
			}
		}
	}

	/**
	 * Calls to the subgroups, to those specified in the paramater coordinations
	 * This map is indexed of the Subgroup ids (IntWrapper) and the values are the
	 * SubgroupCoordinations.
	 * If the parameter handleNullCoordinatios is true, every subgroup is called,
	 * even if the associated coordination is null or shows the end of the transfer.
	 * If is false, only the subgroups with valid phases are called.
	 * When the subgroup is called, the coordination obtained is stored in the same
	 * map
	 * @returns true if any of the coordination is valid at the end of the call
	 */
	void callSubgroups(StateTransferTask task, CoordinationsMap coordinations, boolean handleNullCoordinations) throws SubgroupException {
		Trace.code("SubgroupsManager.java -> void callSubgroups ( StateTransferTask task , CoordinationsMap coordinations , boolean handleNullCoordinations )");
		if (handleNullCoordinations || coordinations.anyValidCoordination) {
			coordinations.resetValidity();
			Iterator it = coordinations.coordinationsMap.entrySet().iterator();
			while (it.hasNext()) {
				Map.Entry entry = (Map.Entry) (it.next());
				IntWrapper subgroupId = (IntWrapper) (entry.getKey());
				SubgroupInfo subgroupInfo = (SubgroupInfo) subgroups.get(subgroupId);
				if (subgroupInfo != null) //perhaps it has been unregistered!
					{
					SubgroupCoordination subgroupCoordination = (SubgroupCoordination) (entry.getValue());
					assert subgroupId.i == subgroupCoordination.id;
					if ((handleNullCoordinations || coordinations.isValidSubgroupCoordination(subgroupCoordination.phase))) {
						try {
							subgroupCoordination.phase = task.call(subgroupId, subgroupInfo.subgroup, subgroupCoordination.phase);
						}
						catch (IOException ex) {
							throw new SubgroupException(subgroupId.i, ex);
						}
						catch (RuntimeException ex) {
							throw new SubgroupException(subgroupId.i, ex);
						}
						coordinations.updateValidity(subgroupCoordination.phase);
					}
				}
			}
		}
		else
			assert !coordinations.getMapState();
	}

	//*************************************************************************************//
	//**************************** INNER CLASS ********************************************//
	//*************************************************************************************//

	class SubgroupInfo {
		public SubgroupInfo(STGroupMember subgroup, DynamicSubgroupInfo info) {
			this.subgroup = subgroup;
			this.info = info;
		}
		public STGroupMember subgroup;
		public DynamicSubgroupInfo info;
	}

	//*************************************************************************************//
	//**************************** DATA MEMBERS *******************************************//
	//*************************************************************************************//

	Map existingGroups; //just to speed up getGroupMemberId!: [GroupMember, IntWrapper]
	SortedMap subgroups; //subgroups defined: IntWrapper, SubgroupInfo, ordered by int
	SubgroupProperties subgroupProperties[]; //it is a cache, to avoid their calculation continuosly
	int singleArray[];
	IntWrapper wrapper;
}
