/**
  * File: GroupMembershipBasicServiceImpl.java
  * Content: Implementation sensei-based of a GroupMembershipService
  * Author: LuisM Pena
  * Date: 5th December 2000
  * Version: 0.29.00
  * Last change:
  *
  **/

package sensei.gmns;

import sensei.middleware.gmns.GroupHandlerFactory;
import sensei.middleware.gmns.GroupHandlerFactoryCreator;
import sensei.middleware.gmns.GroupHandlerFactoryException;
import sensei.middleware.gmns.GroupMembershipBasicServiceBaseImpl;
import sensei.middleware.gmns.InvalidGroupHandlerException;
import sensei.middleware.gmns.InvalidGroupHandlerFactoryException;
import sensei.middleware.gms.GMSnarrower;
import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.SenseiGMSMember;

/**
  * Implementation sensei-based of a GroupMembershipService.
  * It defines static methods as well as the GroupMembershipBasicService interface, to
  *  be used as utility class.
  **/
public class GroupMembershipBasicServiceImpl extends GroupMembershipBasicServiceBaseImpl {

	public GroupMembershipBasicServiceImpl() throws Exception {
		Trace.code("GroupMembershipBasicServiceImpl.java -> GroupMembershipBasicServiceImpl ( )");
	}

	//*************************************************************************************//
	//**************************** GET FACTORY CREATOR ************************************//
	//*************************************************************************************//

	/**
	  * Creates a GroupHandlerFactoryCreator
	  **/
	public GroupHandlerFactoryCreator getFactoryCreator() {
		Trace.code("GroupMembershipBasicServiceImpl.java -> GroupHandlerFactoryCreator getFactoryCreator ( )");
		return utilGetFactoryCreator();
	}

	//*************************************************************************************//
	//**************************** CREATE GROUP *******************************************//
	//*************************************************************************************//

	/**
	  * Creates a new group. Returns the groupHandler created, that
	  * can be invalid if the group cannot be created.
	  * Under error conditions, null is also returned and the error catched through
	  * the error class
	  **/
	public GroupHandler createGroup(GroupHandlerFactory factory) throws InvalidGroupHandlerFactoryException, GroupHandlerFactoryException {
		Trace.code("GroupMembershipBasicServiceImpl.java -> GroupHandler createGroup ( GroupHandlerFactory factory )");
		return utilCreateGroup(factory);
	}

	//*************************************************************************************//
	//**************************** JOIN GROUP *********************************************//
	//*************************************************************************************//

	/**
	  * Joins the member to an existing group. This call doesn't
	  * return until the member joins the group or until the
	  * group handler used becomes invalid (because it belongs to
	  * a member that is expulsed ot leaves its group). In that case, a null reference
	  * is returned
	  * Under error conditions, null is also returned and the error catched through
	  * the error class
	  **/
	public GroupHandler joinGroup(GroupHandler group, GroupHandlerFactory factory) throws InvalidGroupHandlerException, InvalidGroupHandlerFactoryException, GroupHandlerFactoryException {
		Trace.code("GroupMembershipBasicServiceImpl.java -> GroupHandler joinGroup ( GroupHandler group , GroupHandlerFactory factory )");
		return utilJoinGroup(group, factory);
	}

	//*************************************************************************************//
	//**************************** UTIL GET FACTORY CREATOR *******************************//
	//*************************************************************************************//

	/**
	  * Creates a GroupHandlerFactoryCreator
	  **/
	public static GroupHandlerFactoryCreator utilGetFactoryCreator() {
		Trace.code("GroupMembershipBasicServiceImpl.java -> GroupHandlerFactoryCreator utilGetFactoryCreator ( )");
		return new GroupHandlerFactoryCreatorImpl();
	}

	//*************************************************************************************//
	//**************************** UTIL CREATE GROUP **************************************//
	//*************************************************************************************//

	/**
	  * Creates a new group. Returns the groupHandler created, that
	  * can be invalid if the group cannot be created
	  **/
	public static GroupHandler utilCreateGroup(GroupHandlerFactory factory) throws InvalidGroupHandlerFactoryException, GroupHandlerFactoryException {
		Trace.code("GroupMembershipBasicServiceImpl.java -> GroupHandler utilCreateGroup ( GroupHandlerFactory factory )");
		SenseiGMSMember member = getSenseiGMSMember(factory);

		try {
			member.createGroup();
		}
		catch (Exception ex) {
			throw new InvalidGroupHandlerFactoryException(ex.getMessage());
		}

		return member;
	}

	//*************************************************************************************//
	//**************************** UTIL JOIN GROUP ****************************************//
	//*************************************************************************************//

	/**
	  * Joins the member to an existing group. This call doesn't
	  * return until the member joins the group or until the
	  * group handler used becomes invalid (because it belongs to
	  * a member that is expulsed ot leaves its group). In that case, a null reference
	  * is returned
	  **/
	public static GroupHandler utilJoinGroup(Object groupHandler, GroupHandlerFactory factory) throws InvalidGroupHandlerException, InvalidGroupHandlerFactoryException, GroupHandlerFactoryException {
		Trace.code("GroupMembershipBasicServiceImpl.java -> GroupHandler utilJoinGroup ( Object groupHandler , GroupHandlerFactory factory )");
		SenseiGMSMember member = getSenseiGMSMember(factory);

		//it must be a SenseiGMSMember!
		SenseiGMSMember castedGroup = GMSnarrower.toSenseiGMSMember(groupHandler);

		if (castedGroup == null) {
			throw new InvalidGroupHandlerException();
		}

		try {
			if (!member.joinGroup(castedGroup))
				member = null;
		}
		catch (Exception ex) {
			try {
				if (member.isValidGroup())
					throw new InvalidGroupHandlerFactoryException(ex.getMessage());
			}
			catch (Exception ex2) {
			}
		}

		return member;
	}

	//*************************************************************************************//
	//**************************** GET SENSEI GMS MEMBER **********************************//
	//*************************************************************************************//

	/**
	  * Returns a handler from the GroupHandlerFactory
	  **/
	static SenseiGMSMember getSenseiGMSMember(GroupHandlerFactory factory) throws InvalidGroupHandlerFactoryException, GroupHandlerFactoryException {
		Trace.code("GroupMembershipBasicServiceImpl.java -> SenseiGMSMember getSenseiGMSMember ( GroupHandlerFactory factory )");
		if (factory == null)
			throw new InvalidGroupHandlerFactoryException();

		GroupHandler handler = null;
		try {
			handler = factory.create();
		}
		catch (GroupHandlerFactoryException gex) {
			throw gex;
		}
		catch (Exception ex) //remote exception
			{
			throw new GroupHandlerFactoryException(ex.getMessage());
		}

		//it must be a SenseiGMSMember! If not: InvalidGroupHandler, or return just null+trace, or assertion?
		//in this case, referred to the factory
		SenseiGMSMember member = GMSnarrower.toSenseiGMSMember(handler);

		if (member == null) {
			throw new InvalidGroupHandlerFactoryException();
		}

		return member;
	}

}
