/**
  * File: GroupMembershipNamingServiceImpl.java
  * Content: Implementation of the GroupMembershipNamingService
  * Author: LuisM Pena
  * Date: 5th December 2000
  * Version: 0.29.00
  * Last change:
  *
  **/

package sensei.gmns;

import sensei.domains.DomainGroupHandlerImpl;
import sensei.domains.GroupMonitorImpl;
import sensei.domains.utils.ExitHandler;
import sensei.middleware.domains.BehaviourOnViewChanges;
import sensei.middleware.domains.DomainGroupHandler;
import sensei.middleware.domains.DynamicSubgroupInfo;
import sensei.middleware.domains.DynamicSubgroupsUserBaseImpl;
import sensei.middleware.domains.GroupMonitor;
import sensei.middleware.domains.MemberStateException;
import sensei.middleware.gmns.GroupHandlerFactory;
import sensei.middleware.gmns.GroupHandlerFactoryCreator;
import sensei.middleware.gmns.GroupHandlerFactoryException;
import sensei.middleware.gmns.GroupMembershipNamingServiceBaseImpl;
import sensei.middleware.gmns.InvalidGroupHandlerFactoryException;
import sensei.middleware.gmns.MapStringSet;
import sensei.middleware.gmns.MapStringSetObserver;
import sensei.middleware.gmns.MemberInfo;
import sensei.middleware.gmns.ReplicatedServer;
import sensei.middleware.gmns.SetMemberInfo;
import sensei.middleware.gms.GroupHandler;
import sensei.middleware.gms.GroupMember;
import sensei.util.SenseiError;

/**
  * Implementation of the GroupMemberNamingService
  **/
public class GroupMembershipNamingServiceImpl extends GroupMembershipNamingServiceBaseImpl {

	/***********************************************************************************************/
	/********************** CONSTRUCTOR ************************************************************/
	/***********************************************************************************************/

	/**
	  * The constructor receives directly the properties to use.
	  * It also requires an object to handle the exclusion from the member from the group.
	  * @exception GMNSexistingException is thrown if a valid GMNS exists already for the
	  * specified configuration file
	  **/
	public GroupMembershipNamingServiceImpl(MapStringSetObserver observer, GMNSAutoObserver gmnsObserver, ExitHandler exitHandler) throws Exception {
		Trace.code("GroupMembershipNamingServiceImpl.java -> GroupMembershipNamingServiceImpl ( MapStringSetObserver observer , GMNSAutoObserver gmnsObserver , ExitHandler exitHandler )");
		initComponents(observer, gmnsObserver, exitHandler);

		//conform a group by itself
		new AutoGMNS(domainHandler, theGroupMembershipNamingService());
	}

	void initComponents(MapStringSetObserver observer, GMNSAutoObserver gmnsObserver, ExitHandler exitHandler) throws Exception {
		Trace.code("GroupMembershipNamingServiceImpl.java -> void initComponents ( MapStringSetObserver observer , GMNSAutoObserver gmnsObserver , ExitHandler exitHandler )");
		//Create first a domainGroupHandler, as different subgroups are to be used
		domainHandler = new DomainGroupHandlerImpl().theDomainGroupHandler();
		domainHandler.setBehaviourMode(BehaviourOnViewChanges.MembersOnTransferExcludedFromGroup);
		domainHandler.setDynamicSubgroupsUser(new MyDynamicSubgroupsUser().theDynamicSubgroupsUser());

		//creates the monitor and the map
		GroupMonitorImpl monitorImpl = new GroupMonitorImpl(Consts.MONITOR_SUBGROUP, domainHandler);
		MapStringSetImpl mapImpl = new MapStringSetImpl(Consts.MAP_SUBGROUP, domainHandler);
		map = mapImpl.theMapStringSet();

		domainHandler.setDomainGroupUser(new DomainGroupHandlerFactory(domainHandler, map, gmnsObserver, exitHandler).theDomainGroupUser());

		monitorImpl.register();
		mapImpl.register();

		//Register the monitor and the map
		monitor = monitorImpl.theGroupMonitor();
		map.addObserver(observer);

		//create the checker (it registers itself)
		new CheckerActivator(domainHandler, map, gmnsObserver);

		creator = new GroupCreator(domainHandler, map, monitor);
		joiner = new GroupJoiner(domainHandler, map);
	}

	//*************************************************************************************//
	//**************************** GET FACTORY CREATOR ************************************//
	//*************************************************************************************//

	/**
	  * Creates a GroupHandlerFactoryCreator
	  **/
	public GroupHandlerFactoryCreator getFactoryCreator() {
		Trace.code("GroupMembershipNamingServiceImpl.java -> GroupHandlerFactoryCreator getFactoryCreator ( )");
		return GroupMembershipBasicServiceImpl.utilGetFactoryCreator();
	}

	/***********************************************************************************************/
	/********************** FIND AND JOIN GROUP ****************************************************/
	/***********************************************************************************************/

	/**
	  * Finds a group with the name specified and includes the member with the given name
	  **/
	public GroupHandler findAndJoinGroup(String groupName, GroupHandlerFactory factory, String memberId, ReplicatedServer reference) throws InvalidGroupHandlerFactoryException, GroupHandlerFactoryException, MemberStateException {
		Trace.code("GroupMembershipNamingServiceImpl.java -> GroupHandler findAndJoinGroup ( String groupName , GroupHandlerFactory factory , String memberId , ReplicatedServer reference )");
		GroupHandler ret = null;
		try {
			while (ret == null) {
				ret = joiner.joinGroup(groupName, factory, memberId, reference);
				if (ret == null)
					if (joiner.errorDueToClient())
						break;
					else {
						ret = creator.createGroup(groupName, factory, memberId, reference);
						if (ret == null && creator.errorDueToClient())
							break;
					}
			}
		}
		catch (MemberStateException ex) {
			throw ex;
		}
		catch (InvalidGroupHandlerFactoryException ex) {
			throw ex;
		}
		catch (GroupHandlerFactoryException ex) {
			throw ex;
		}
		catch (Exception ex) {
			SenseiError.unhandledException(Consts.AREA, ex);
		}
		return ret;
	}

	/***********************************************************************************************/
	/********************** LEAVE GROUP ************************************************************/
	/***********************************************************************************************/

	/**
	  * Leaves a group. This method is not needed to be called, as the service can
	  * discover by itself that a member has left the group, but it can speed up
	  * other member's joining.
	  **/
	public void leaveGroup(String groupName, GroupHandler member) {
		Trace.code("GroupMembershipNamingServiceImpl.java -> void leaveGroup ( String groupName , GroupHandler member )");
		try {
			SetMemberInfo group = map.get(groupName);
			if (group != null) {
				group.remove(new MemberInfo(member, "", null));
				//try to remove the group, if it's empty
				if (group.get().handler == null)
					removeGroup(groupName, group);
			}
		}
		catch (Exception ex) {
		}
	}

	void removeGroup(String groupName, SetMemberInfo group) {
		Trace.code("GroupMembershipNamingServiceImpl.java -> void removeGroup ( String groupName , SetMemberInfo group )");
		try {
			domainHandler.startTransaction(monitor);
			if (group.get().handler == null) {
				map.remove(groupName);
				domainHandler.removeSubgroup(domainHandler.getSubgroupId(group), Consts.noInfo);
			}
		}
		catch (Exception ex) {
		}
		try {
			domainHandler.endTransaction();
		}
		catch (Exception ex) {
		}
	}

	/***********************************************************************************************/
	/********************** GET GROUP **************************************************************/
	/***********************************************************************************************/

	public ReplicatedServer[] getGroup(String name) throws MemberStateException {
		Trace.code("GroupMembershipNamingServiceImpl.java -> ReplicatedServer [ ] getGroup ( String name )");
		ReplicatedServer ret[] = null;
		try {
			SetMemberInfo set = map.get(name);
			if (set == null)
				ret = new ReplicatedServer[0];
			else {
				MemberInfo info[] = set.toArray();
				int size = info.length;
				int retSize = 0;
				for (int i = 0; i < size; i++)
					if (info[i].publicReference != null)
						retSize++;
				ret = new ReplicatedServer[retSize];
				for (int i = 0; i < size; i++) {
					ReplicatedServer toUse = info[i].publicReference;
					if (toUse != null)
						ret[--retSize] = toUse;
				}
			}
		}
		catch (MemberStateException mex) {
			throw mex;
		}
		catch (Exception ex) {
			SenseiError.unhandledException(Consts.AREA, ex);
		}
		return ret;
	}

	/***********************************************************************************************/
	/********************** GET GROUP SERVER *******************************************************/
	/***********************************************************************************************/

	public ReplicatedServer getGroupServer(String name, String member) throws MemberStateException {
		Trace.code("GroupMembershipNamingServiceImpl.java -> ReplicatedServer getGroupServer ( String name , String member )");
		ReplicatedServer ret = null;
		try {
			SetMemberInfo set = map.get(name);
			if (set != null) {
				MemberInfo info[] = set.toArray();
				int size = info.length;
				for (int i = 0; i < size; i++)
					if (member.equals(info[i].name)) {
						ret = info[i].publicReference;
						break;
					}
			}
		}
		catch (MemberStateException mex) {
			throw mex;
		}
		catch (Exception ex) {
			SenseiError.unhandledException(Consts.AREA, ex);
		}
		return ret;
	}

	/***********************************************************************************************/
	/********************** GET VALID GROUP SERVER *************************************************/
	/***********************************************************************************************/

	public ReplicatedServer getValidGroupServer(String name) throws MemberStateException {
		Trace.code("GroupMembershipNamingServiceImpl.java -> ReplicatedServer getValidGroupServer ( String name )");
		ReplicatedServer ret = null;
		try {
			SetMemberInfo set = map.get(name);
			if (set != null) {
				MemberInfo info[] = set.toArray();
				int size = info.length;
				for (int i = 0; i < size; i++)
					try {
						if (info[i].handler.isValidGroup()) {
							ret = info[i].publicReference;
							if (ret != null)
								break;
						}
					}
					catch (Exception ex) {
					}
			}
		}
		catch (MemberStateException mex) {
			throw mex;
		}
		catch (Exception ex) {
			SenseiError.unhandledException(Consts.AREA, ex);
		}
		return ret;
	}

	//*************************************************************************************//
	//**************************** DYNAMIC SUBGROUPS USER *********************************//
	//*************************************************************************************//

	class MyDynamicSubgroupsUser extends DynamicSubgroupsUserBaseImpl {
		public MyDynamicSubgroupsUser() throws Exception {
		}
		public GroupMember acceptSubgroup(int id, DynamicSubgroupInfo info) {
			try {
				return new SetMemberInfoImpl(id, domainHandler).theSetMemberInfo();
			}
			catch (Exception ex) {
				return null;
			}
		}
		public GroupMember subgroupCreated(int creator, int id, DynamicSubgroupInfo info) {
			try {
				return new SetMemberInfoImpl(id, domainHandler).theSetMemberInfo();
			}
			catch (Exception ex) {
				return null;
			}
		}
		public void subgroupRemoved(int remover, int id, DynamicSubgroupInfo info) {
		}
	}

	//*************************************************************************************//
	//**************************** DATA MEMBERS *******************************************//
	//*************************************************************************************//

	DomainGroupHandler domainHandler;
	GroupMonitor monitor;
	MapStringSet map;
	GroupCreator creator;
	GroupJoiner joiner;
}
