/**
  * File: CommunicationChannel.java
  * Content: asynchronous communication channel between two SenseiGMSMember objects
  * Author: LuisM Pena
  * Date: 3nd July 2000
  * Version: 0.97.00
  * Last change:
  **/

package sensei.gms;

import sensei.middleware.gms.SenseiGMSMember;

import sensei.util.SenseiError;
import sensei.util.RestartableThread;

/**
  * Asynchronous communication channel between two SenseiGMSMember objects
  **/
class CommunicationChannel extends RestartableThread {
	/**
	  * A communication channel needs always to know the SenseiGMSMember
	  * to link. The asynchronous behaviour is got by using threads,
	  * therefore a ThreadGroup is needed to group the different
	  * communication channels
	  * @param link the SenseiGMSMember to communicate
	  * @param linkId the SenseiGMSMember's group id, used just to trace info.
	  * @param threadGroup a ThreadGroup to link every communication
	  * channel belonging to a specific CORBA object
	  **/
	public CommunicationChannel(SenseiGMSMember link, int linkId, ThreadGroup threadGroup) {
		super(threadGroup, THREAD_NAME + linkId);
		Trace.code("CommunicationChannel.java -> CommunicationChannel ( SenseiGMSMember link , int linkId , ThreadGroup threadGroup )");
		this.link = link;
		this.linkId = linkId;
		fineState = true;
	}

	//*************************************************************************************//
	//**************************** FREE RESOURCES *****************************************//
	//*************************************************************************************//

	/**
	  * Releases any resource associated to this object (~destructor).
	  * This method MUST be always called, as otherwise the associated thread will run forever
	  **/
	public void freeResources() {
		Trace.code("CommunicationChannel.java -> void freeResources ( )");
		if (isReliable()) {
			Trace.releaseObject("CommunicationChannel to " + linkId);

			fineState = false;
			end(); //finish the thread
			link = null;

		}
	}

	//*************************************************************************************//
	//**************************** IS RELIABLE ********************************************//
	//*************************************************************************************//

	/**
	  * Returns true if the channel is reliable, has no failures
	  **/
	public final boolean isReliable() {
		Trace.code("CommunicationChannel.java -> final boolean isReliable ( )");
		return fineState;
	}

	//*************************************************************************************//
	//**************************** GET LINK ***********************************************//
	//*************************************************************************************//

	/**
	  * Returns the SenseiGMSMember that this channel links
	  **/
	public SenseiGMSMember getLink() {
		Trace.code("CommunicationChannel.java -> SenseiGMSMember getLink ( )");
		return link;
	}

	//*************************************************************************************//
	//**************************** GET LINK ID ********************************************//
	//*************************************************************************************//

	/**
	  * Returns the SenseiGMSMember's GroupMemberId that this channel links
	  **/
	public int getLinkId() {
		Trace.code("CommunicationChannel.java -> int getLinkId ( )");
		return linkId;
	}

	//*************************************************************************************//
	//**************************** INVOKE *************************************************//
	//*************************************************************************************//

	/**
	  * Does a request asynchronously.
	  * After this call, the requester will receive a call to invokeMember on the
	  * member associated to this channel.
	  * @return true if the request is accepted
	  **/
	public boolean invoke(ChannelRequester requester) {
		Trace.code("CommunicationChannel.java -> boolean invoke ( ChannelRequester requester )");
		if (isReliable()) {
			try {
				if (!activate(CHANNEL_DEAD, requester)) {
					invalidateObject();
				}
			}
			catch (InterruptedException ex) {
				invalidateObject();
				SenseiError.unhandledException(Consts.GMS, ex);
			}
		}
		return isReliable();
	}

	//*************************************************************************************//
	//**************************** WAIT CHANNEL ****************************************//
	//*************************************************************************************//

	/**
	  * Waits for the completion of a previous request
	  * After this call, the requester will receive a call to invokeMember on the
	  * member associated to this channel.
	  * @return true if the request is accepted
	  **/
	public boolean waitChannel() {
		Trace.code("CommunicationChannel.java -> boolean waitChannel ( )");
		if (isReliable()) {
			try {
				if (!waitCurrent(CHANNEL_DEAD)) {
					invalidateObject();
				}
			}
			catch (InterruptedException ex) {
				invalidateObject();
				SenseiError.unhandledException(Consts.GMS, ex);
			}
		}
		return isReliable();
	}

	//*************************************************************************************//
	//**************************** ACTIVATED **********************************************//
	//*************************************************************************************//

	/**
	  * RestartableThread method, is the real communication center to the link
	  * @param taskParam must be a ChannelRequester
	  **/
	public final boolean activated(Object taskParam) {
		Trace.code("CommunicationChannel.java -> final boolean activated ( Object taskParam )");
		ChannelRequester requester = (ChannelRequester) taskParam;
		return requester.invokeMember(this);
	}

	//*************************************************************************************//
	//**************************** INVALIDATE OBJECT **************************************//
	//*************************************************************************************//

	/**
	  * Invalidates the object
	  **/
	void invalidateObject() {
		Trace.code("CommunicationChannel.java -> void invalidateObject ( )");
		freeResources();
	}

	//*************************************************************************************//
	//**************************** DATA MEMBERS *******************************************//
	//*************************************************************************************//

	SenseiGMSMember link; //the SenseiGMSMember to communicate
	int linkId; //its GroupMemberId

	boolean fineState; //set to false if the channel can't be used anymore

	static long channelNumber = 0; //used to create thread names

	static final String THREAD_NAME = "CommunicationChannel.";
	//CHANNEL_DEAD contains the maximum time a channel waits for an answer,
	// once expired the channel is considered dead
	static final int CHANNEL_DEAD = Consts.CHANNEL_DEAD;
}
