/**
  * File: Trace.java
  * Content: Class to trace the GMS events
  * Author: LuisM Pena
  * Date: 9th August 2000
  * Last change:
  * Version: 0.97.00
  *
  * Comments: This class is not intended to be included in the release
  *						version.
  **/

package sensei.gms;

import java.util.Arrays;
import java.util.Vector;

import sensei.middleware.gms.InternalView;
import sensei.middleware.gms.InternalViewId;
import sensei.middleware.gms.MessageId;
import sensei.util.SenseiError;
import sensei.util.logging.Logger;

/**
  * TraceGraphical defines the base interface to log the activity
  * of GroupMembers.
  **/
public class Trace {
	/**
	  * Add a consumer of the trace events
	  * Because this operation is not a often used one, its implementation
	  * is an slow one, speeding up the access to the tracers later
	  **/
	public static void addTracer(TraceGraphical tracer) {
		if (TRACE)
			if (nTracers == 0) {
				tracers = new TraceGraphical[1];
				tracers[0] = tracer;
				nTracers = 1;
			}
			else {
				Vector listTracers = new Vector(Arrays.asList(tracers));
				listTracers.add(tracer);
				nTracers = listTracers.size();
				tracers = new TraceGraphical[nTracers];
				for (int i = 0; i < nTracers; i++)
					tracers[i] = (TraceGraphical) (listTracers.elementAt(i));
			}
	}

	public static void code(String code) {
		if (TRACE_CODE) {
			printAlways("Code:" + code);
		}
	}

	public static void memberTooSlow() {
		if (TRACE) {
			printAlways("Member too slow");
			for (int i = 0; i < nTracers; i++)
				tracers[i].memberTooSlow();
		}
	}

	public static void memberExitGroup() {
		if (TRACE) {
			printAlways("Member excluded from the group");
			for (int i = 0; i < nTracers; i++)
				tracers[i].memberExitGroup();
		}
	}

	public static void token(boolean received) {
		if (TRACE_TOKEN) {
			if (received)
				printAlways("Token received");
			else
				printAlways("Token released");
			for (int i = 0; i < nTracers; i++)
				tracers[i].token(received);
		}
	}

	public static void viewReceived(InternalView view) {
		if (TRACE)
			printMembersAlways("InternalView received: " + convertViewId(view.viewId), view.memberIds);
	}

	public static void wrongView(InternalView view, String reason) {
		if (TRACE)
			printMembersAlways("Wrong InternalView received: [" + reason + "]: " + convertViewId(view.viewId), view.memberIds);
	}

	public static void viewAccepted(InternalView view, int member) {
		if (TRACE) {
			int rank = getRank(view, member);
			validView("InternalView accepted: ", view, rank);
			for (int i = 0; i < nTracers; i++)
				tracers[i].viewAccepted(view, rank);
		}
	}

	public static void changingViewEventSent() {
		if (TRACE)
			printAlways("Changing view event sent");
	}

	public static void viewSent(InternalView view) {
		if (TRACE)
			printMembersAlways("InternalView sent: " + convertViewId(view.viewId), view.memberIds);
	}

	public static void viewInstalled(InternalView view, int member) {
		if (TRACE) {
			int rank = getRank(view, member);
			validView("InternalView installed: ", view, rank);
			for (int i = 0; i < nTracers; i++)
				tracers[i].viewInstalled(view, rank);
		}
	}

	public static void viewExclusionSent(InternalView view) {
		if (TRACE)
			printMembersAlways("InternalView with autoexclusion sent: " + convertViewId(view.viewId), view.memberIds);
	}

	public static void messageReceived(MessageId messageId, boolean confirmation) {
		if (TRACE) {
			printAlways((confirmation ? "Message confirmed: " : "Message received: ") + convertMessageId(messageId));
			for (int i = 0; i < nTracers; i++)
				tracers[i].messageReceived(messageId, confirmation);
		}
	}

	public static void messageSent(MessageId messageId, boolean confirmation) {
		if (TRACE)
			printAlways((confirmation ? "Message confirmation sent: " : "Message sent: ") + convertMessageId(messageId));
	}

	public static void messagePTPReceived(int sender) {
		if (TRACE)
			printAlways("PTP Message received from " + sender);
	}

	public static void messagePTPSent(int target) {
		if (TRACE)
			printAlways("PTP Message sent to " + target);
	}

	public static final int TokenRecoveryAsking = 0;
	public static final int TokenRecoveryDropped = 1;
	public static final int TokenRecoveryStarted = 2;
	public static void tokenRecovery(int status) {
		if (TRACE) {
			if (status == 0)
				printAlways("Starts the request to try to recover the token");
			else if (status == 1)
				printAlways("Quitting the token recovery");
			else if (status == 2)
				printAlways("Recovering token");
		}
	}

	//  public static void requestedTokenRecovery(int requester, AllowTokenRecoveryAnswer answer)
	//  {
	//    if (TRACE)
	//    {
	//      if (answer.validCommunication)
	//        printAlways(requester+" has requested token, reply is "+answer.recoveryAllowed);
	//      else
	//        printAlways(requester+" has requested token, but it's not a right member");
	//    }
	//  }
	//
	public static void requestedTokenRecovery(int requester, boolean answer, String reason) {
		if (TRACE) {
			printAlways(requester + " has requested token, reply is " + (answer ? "OK" : "Not Ok") + " beacuse: " + reason);
		}
	}

	public static void wrongMessage(int member, MessageId messageId, boolean onConfirmation) {
		if (TRACE) {
			if (onConfirmation)
				printAlways(member + " has sent a wrong confirmation message, with id=" + convertMessageId(messageId));
			else
				printAlways(member + " has sent a wrong message, with id=" + convertMessageId(messageId));
		}
	}

	public static void wrongCommunication(int toMember, int step, String operation) {
		if (TRACE)
			printAlways("Failed communication (" + operation + ") by " + (step == 1 ? "invocation" : "delayed communication") + " to " + toMember);
	}

	public static void groupCommunication(int sender, String operation, boolean accepted) {
		if (TRACE)
			printAlways("Communication " + operation + " from " + sender + " completed .Accepted=" + accepted);
	}

	public static void print(String message) {
		if (TRACE) {
			printAlways(message);
			for (int i = 0; i < nTracers; i++)
				tracers[i].printAlways(message);
		}
	}

	public static void printMembers(String message, int members[]) {
		if (TRACE) {
			printMembersAlways(message, members);
			for (int i = 0; i < nTracers; i++)
				tracers[i].printMembers(message, members);
		}
	}

	public static void releaseObject(String object) {
		if (TRACE)
			printAlways("Released object: " + object);
	}

	public static void handledException(Exception ex) {
		if (TRACE_EXCEPTIONS) {
			String s;
			if (TRACE_EXCEPTIONS_LONG_FORMAT)
				s = SenseiError.convertExceptionToString(ex);
			else
				s = ex.toString();
			printAlways(s);
		}
	}

	static void printAlways(String message) {
		if (logger != null)
			logger.log(message);
	}

	static void printMembersAlways(String message, int memberIds[]) {
		StringBuffer s = new StringBuffer(message + " Members : " + memberIds.length + " :[ ");
		for (int i = 0; i < memberIds.length; i++) {
			if (i > 0)
				s.append(" - ");
			s.append(memberIds[i]);
		}
		s.append(" ]");
		printAlways(s.toString());
	}

	static int getRank(InternalView view, int member) {
		int members[] = view.memberIds;
		int i = members.length;
		while (i-- > 0)
			if (members[i] == member)
				break;
		return i;
	}

	static void validView(String s, InternalView view, int rank) {
		printMembersAlways(s + convertViewId(view.viewId) + "-> my rank is " + rank + "; ", view.memberIds);
	}

	static String convertMessageId(MessageId messageId) {
		return "(" + messageId.view + "," + messageId.id + ")";
	}

	static String convertViewId(InternalViewId viewId) {
		return "(" + viewId.id + "," + viewId.installing + ")";
	}

	static TraceGraphical[] tracers = null;
	static int nTracers = 0;

	static public final boolean TRACE = sensei.util.Configuration.getSingleton().traceGMS();
	static public final boolean TRACE_CODE = sensei.util.Configuration.getSingleton().traceGMSCode();
	static public final boolean TRACE_TOKEN = sensei.util.Configuration.getSingleton().traceGMSToken();
	static public final boolean TRACE_EXCEPTIONS = sensei.util.Configuration.getSingleton().traceExceptionsGMS();
	static public final boolean TRACE_EXCEPTIONS_LONG_FORMAT = sensei.util.Configuration.getSingleton().traceExceptionsDetailedGMS();
	static final Logger logger = Logger.getLogger(Consts.GMS, TRACE);
}
