/**
  * File: SenseiError.java
  * Content: Basic error facilities
  * Author: LuisM Pena
  * Date: 19th August 2000
  * Version: 0.97.00
  * Last change:
  *
  *
  **/

package sensei.util;

import java.io.PrintWriter;
import java.io.StringWriter;

import sensei.util.logging.Logger;

/**
  * SenseiError is a utility class which allows the code to redirect errors easily.
  * It has two main methods, error and unhandledException, to show the error conditions.
  * The output can be controlled in three ways. By default, it is sent to the System.err,
  * but this can be avoided with useStandardOutput method. It can use as well the Logger,
  * with useLogger method, and finally, it can be redirected to some SenseiError catcher implementing
  * the ErrorHandler interface.
  **/
public class SenseiError {
	/**
	  * Defines wether to use the standard output error to show errors. By default, this
	  * is set to true
	  **/
	public static void useStandardOutput(boolean use) {
		SenseiError.useErr = use;
	}

	/**
	  * Sets the Logger to receive the debug messages. (By default is true)
	  **/
	public static void useLogger(boolean use) {
		SenseiError.useLog = use;
	}

	/**
	  * Sets the output to receive the debug messages. Returns the previous output installed.
	  * If there is an error handler
	  **/
	public static ErrorHandler setErrorHandler(ErrorHandler handler) {
		ErrorHandler old = SenseiError.handler;
		SenseiError.handler = handler;
		return old;
	}

	/**
	  * raises a verified error
	  **/
	public static void error(String area, String reason) {
		boolean handled = false;
		if (handler != null)
			handled = handler.handleError(area, reason);
		if (!handled) {
			String error = "Execution error [" + area + "]: " + reason;
			message(area, error);
			throw new RuntimeException(error);
		}
	}

	/**
	  * Handles unexpected exceptions
	  **/
	public static void unhandledException(String area, Exception ex) {
		boolean handled = false;
		//no reason to throw again now the exception!!!!
		if (handler != null)
			handled = handler.handleException(area, ex);
		if (!handled)
			message(area, "UnhandledException [" + area + "]: " + convertExceptionToString(ex));
	}

	public static String convertExceptionToString(Exception ex) {
		StringWriter sw = new StringWriter();
		ex.printStackTrace(new PrintWriter(sw));
		return sw.toString();
	}

	static void message(String area, String s) {
		if (useErr)
			System.err.println(s);

		if (useLog) {
			Logger logger = Logger.getLogger(area, true);
			if (logger != null)
				logger.log(s);
		}
	}

	static ErrorHandler handler = null;
	static boolean useErr = true;
	static boolean useLog = true;
}
